%name_prefix="textrep_"
%pure-parser
%defines
%{
#define YYDEBUG 1
#include "TextRepScanClass.h"
#include "plstr.h"

#define MATHML_NS NS_LITERAL_STRING("http://www.w3.org/1998/Math/MathML").get()

#define MAKE_MATHML_OBJECT(store,storetype,eltype) \
nsCOMPtr<mathml_dom_I##storetype> store; \
{\
  nsCOMPtr<dom_IElement> _mmo_tmpstore; \
  nsresult _mmo_ret = \
    aDocument->CreateElementNS(MATHML_NS, NS_LITERAL_STRING(eltype).get(), \
                               getter_AddRefs(_mmo_tmpstore)); \
  NS_ENSURE_SUCCESS(_mmo_ret, _mmo_ret); \
  store = do_QueryInterface(_mmo_tmpstore, &_mmo_ret); \
  NS_ENSURE_SUCCESS(_mmo_ret, _mmo_ret); \
}

struct PropertyApplicator
{
  const char* propertyName;
  void (*applyProperty)(mathml_dom_IMathMLContentElement* aEl, nsCString& aTarget);
};

#define DEFINE_PROPERTYAPPLICATOR(castTo, objName) \
void apply##objName(mathml_dom_IMathMLContentElement* aEl, nsCString& aTarget) \
{ \
  NS_ConvertUTF8toUTF16 t(aTarget); \
  NS_STATIC_CAST(mathml_dom_IMathML##castTo*, aEl)->Set##objName(t.get()); \
}

#define DEFINE_PROPERTYAPPLICATOR_AMBIG(castTo, objName) \
void apply##objName##castTo(mathml_dom_IMathMLContentElement* aEl, nsCString& aTarget) \
{ \
  NS_ConvertUTF8toUTF16 t(aTarget); \
  NS_STATIC_CAST(mathml_dom_IMathML##castTo*, aEl)->Set##objName(t.get()); \
}

DEFINE_PROPERTYAPPLICATOR(ContentElement, ClassName);
DEFINE_PROPERTYAPPLICATOR(ContentElement, MathElementStyle);
DEFINE_PROPERTYAPPLICATOR(ContentElement, Id);
DEFINE_PROPERTYAPPLICATOR(ContentElement, Xref);
DEFINE_PROPERTYAPPLICATOR(ContentElement, Href);
DEFINE_PROPERTYAPPLICATOR_AMBIG(ContentToken, DefinitionURL);
DEFINE_PROPERTYAPPLICATOR_AMBIG(ContentToken, Encoding);
DEFINE_PROPERTYAPPLICATOR_AMBIG(CnElement, Type);
DEFINE_PROPERTYAPPLICATOR(CnElement, Base);
DEFINE_PROPERTYAPPLICATOR_AMBIG(CiElement, Type);
DEFINE_PROPERTYAPPLICATOR_AMBIG(PredefinedSymbol, DefinitionURL);
DEFINE_PROPERTYAPPLICATOR_AMBIG(PredefinedSymbol, Encoding);

void
applyUnits(mathml_dom_IMathMLContentElement* aEl, nsCString& aTarget)
{
  NS_ConvertUTF8toUTF16 t(aTarget);
  nsCOMPtr<dom_IDocument> doc;
  aEl->GetOwnerDocument(getter_AddRefs(doc));
  nsCOMPtr<dom_IElement> el;
  doc->GetDocumentElement(getter_AddRefs(el));
  nsString nsURI;
  el->GetNamespaceURI(getter_Copies(nsURI));
  aEl->SetAttributeNS(nsURI.get(), NS_LITERAL_STRING("units").get(), t.get());
}

static PropertyApplicator pat_element[] =
{
  {"class", applyClassName},
  {"style", applyMathElementStyle},
  {"id", applyId},
  {"Xref", applyXref},
  {"Href", applyHref},
  {nsnull, nsnull}
};

static PropertyApplicator pat_token[] =
{
  {"definitionURL", applyDefinitionURLContentToken},
  {"encoding", applyEncodingContentToken},
  {nsnull, nsnull}
};

static PropertyApplicator pat_cn[] =
{
  {"type", applyTypeCnElement},
  {"base", applyBase},
  {"units", applyUnits},
  {nsnull, nsnull}
};

static PropertyApplicator pat_ci[] =
{
  {"type", applyTypeCiElement},
  {nsnull, nsnull}
};

static PropertyApplicator pat_predef[] =
{
  {"definitionURL", applyDefinitionURLPredefinedSymbol},
  {"encoding", applyEncodingPredefinedSymbol},
  {nsnull, nsnull}
};

/* This is used for a range of content elements which don't have any further
 * properties...
 */
static PropertyApplicator* patt_contentelement[] =
{
  pat_element,
  nsnull
};

static PropertyApplicator* patt_cn[] =
{
  pat_element,
  pat_token,
  pat_cn,
  nsnull
};

static PropertyApplicator* patt_ci[] =
{
  pat_element,
  pat_token,
  pat_ci,
  nsnull
};

static PropertyApplicator* patt_predef[] =
{
  pat_element,
  pat_predef,
  nsnull
};

static void
textrep_ApplyPropertyList(mathml_dom_IMathMLContentElement* aEl,
                          PropertyApplicator** aPropTables,
                          nsCStringArray& aProps)
{
  PropertyApplicator ** p = aPropTables, * q;
  PRInt32 i, l = aProps.Count();
  for (i = 0; i < l - 1; i += 2)
  {
    for (q = *p; (q = *p); p++)
    {
      for (; q->propertyName; q++)
        if ((*aProps[i]) == q->propertyName)
        {
          q->applyProperty(aEl, *aProps[i + 1]);
          break;
        }
      if (q->propertyName)
        break;
    }
  }
}

static bool
textrep_IsConstant(nsCString& aName)
{
  const char* t = aName.get();
  if (t[0] <= 'i')
  {
    if (t[0] <= 'e')
    {
      if (t[0] == 'e')
      {
        if (t[1] == 'm')
          // emptyset
          return strcmp(t + 2, "ptyset") == 0;
        else if (t[1] == 'u')
          // eulergamma
          return strcmp(t + 2, "lergamma") == 0;
        else if (t[1] == 'x')
          // exponentiale
          return strcmp(t + 2, "ponentiale") == 0;
        else
          return PR_FALSE;
      }
      else if (t[0] == 'c')
      {
        // complexes
        return strcmp(t + 1, "omplexes") == 0;
      }
      else
        return PR_FALSE;
    }
    else if (t[0] == 'i')
    {
      if (t[1] == 'n')
      {
        if (t[2] == 'f')
          // infinity
          return strcmp(t + 3, "inity") == 0;
        else if (t[2] == 't')
          // integers
          return strcmp(t + 3, "egers") == 0;
        else
          return PR_FALSE;
      }
      else if (t[1] == 'm')
        // imaginaryi
        return strcmp(t + 2, "aginaryi") == 0;
      else
        return PR_FALSE;
    }
    else if (t[0] == 'f')
      // false
      return strcmp(t + 1, "alse") == 0;
    else
      return PR_FALSE;
  }
  else
  {
    if (t[0] <= 'p')
    {
      if (t[0] == 'n')
      {
        if (t[1] == 'a')
          // naturalnumbers
          return strcmp("turalnumbers", t + 2) == 0;
        else if (t[1] == 'o')
          // notanumber
          return strcmp("tanumber", t + 2) == 0;
        else
          return PR_FALSE;
      }
      else if (t[0] == 'p')
      {
        if (t[1] == 'i')
          // pi
          return t[2] == 0;
        else if (t[1] == 'r')
          // primes
          return strcmp("imes", t + 2) == 0;
        else
          return PR_FALSE;
      }
    }
    else
    {
      if (t[0] == 'r')
      {
        if (t[1] == 'a')
          return strcmp(t + 2, "tionals") == 0;
        else if (t[1] == 'e')
          return strcmp(t + 2, "als") == 0;
        else
          return PR_FALSE;
      }
      else if (t[0] == 't')
        // true
        return strcmp(t + 1, "rue") == 0;
      else
        return PR_FALSE;
    }
  }
  // Unreachable, but gcc complains if this isn't here...
  return PR_FALSE;
}

static nsresult
textrep_MakeCIOrConstant
(
 nsCString& aName,
 bool aOverrideConstant,
 nsCStringArray& aPropList,
 dom_IDocument* aDocument,
 mathml_dom_IMathMLContentElement** aEl
)
{
  if (aOverrideConstant || !textrep_IsConstant(aName))
  {
    MAKE_MATHML_OBJECT(ciEl, MathMLCiElement, "ci");
    nsCOMPtr<dom_IText> tEl;
    NS_ConvertUTF8toUTF16 name(aName);
    nsresult ret =
      aDocument->CreateTextNode(name.get(), getter_AddRefs(tEl));
    NS_ENSURE_SUCCESS(ret, ret);
    nsCOMPtr<dom_INode> discard;
    ret = ciEl->InsertArgument(tEl, 1, getter_AddRefs(discard));
    NS_ENSURE_SUCCESS(ret, ret);

    textrep_ApplyPropertyList(ciEl, patt_ci, aPropList);
    
    *aEl = ciEl;
    NS_ADDREF(*aEl);
    return NS_OK;
  }
  else
  {
    nsCOMPtr<dom_IElement> tmpstore;
    NS_ConvertUTF8toUTF16 name(aName);
    nsresult ret =
      aDocument->CreateElementNS(MATHML_NS, name.get(),
                                 getter_AddRefs(tmpstore));
    NS_ENSURE_SUCCESS(ret, ret);
    nsCOMPtr<mathml_dom_IMathMLContentElement> store =
      do_QueryInterface(tmpstore, &ret);
    NS_ENSURE_SUCCESS(ret, ret);

    textrep_ApplyPropertyList(store, patt_predef, aPropList);

    *aEl = store;
    NS_ADDREF(*aEl);
    
    return NS_OK;
  }
}

static nsresult
textrep_MakeCaseElement
(
 mathml_dom_IMathMLContentElement* aCondition,
 mathml_dom_IMathMLContentElement* aValue,
 dom_IDocument* aDocument,
 nsCStringArray& aPropList, 
 mathml_dom_IMathMLContentElement** aEl
)
{
  MAKE_MATHML_OBJECT(ceEl, MathMLCaseElement, "piece");
  nsresult ret = ceEl->SetCaseValue(aValue);
  NS_ENSURE_SUCCESS(ret, ret);
  ret = ceEl->SetCaseCondition(aCondition);
  NS_ENSURE_SUCCESS(ret, ret);

  textrep_ApplyPropertyList(ceEl, patt_contentelement, aPropList);

  *aEl = ceEl;
  NS_ADDREF(*aEl);
  return NS_OK;
}

static nsresult
textrep_DoInorderExpression
(
 const nsString& aOp, TextRepLValue& aLHS, TextRepLValue& aRHS,
 dom_IDocument* aDocument,
 nsCStringArray& aPropList,
 mathml_dom_IMathMLContentElement** aEl
)
{
  // Firstly, we try to merge into the left-hand expression...
  if (aPropList.Count() == 0)
  {
    nsCOMPtr<mathml_dom_IMathMLApplyElement> apply(do_QueryInterface(aLHS.mMath));
    if (apply != nsnull)
    {
      nsCOMPtr<mathml_dom_IMathMLElement> op;
      apply->GetOperator(getter_AddRefs(op));
      nsString nsURI, lname;
      op->GetNamespaceURI(getter_Copies(nsURI));
      op->GetLocalName(getter_Copies(lname));
      if (nsURI.Equals(MATHML_NS) && lname.Equals(aOp))
      {
        nsCOMPtr<dom_INode> ignore;
        apply->AppendChild(aRHS.mMath, getter_AddRefs(ignore));
        *aEl = apply;
        NS_ADDREF(*aEl);
        return NS_OK;
      }
    }
  }

  // We need a new apply element and operator...
  MAKE_MATHML_OBJECT(apply, MathMLApplyElement, "apply");
  nsCOMPtr<dom_IElement> operatorDOMEl;
  nsresult ret = aDocument->CreateElementNS(MATHML_NS, aOp.get(),
                                            getter_AddRefs(operatorDOMEl));
  NS_ENSURE_SUCCESS(ret, ret);
  nsCOMPtr<mathml_dom_IMathMLContentElement> operatorEl
    (do_QueryInterface(operatorDOMEl, &ret));
  NS_ENSURE_SUCCESS(ret, ret);

  nsCOMPtr<dom_INode> opElClone;
  apply->AppendChild(operatorEl, getter_AddRefs(opElClone));

  // We now have a new element, so add the left and right hand sides...
  apply->AppendChild(aLHS.mMath, getter_AddRefs(opElClone));
  apply->AppendChild(aRHS.mMath, getter_AddRefs(opElClone));
  
  // Put attributes onto the apply...
  textrep_ApplyPropertyList(operatorEl, patt_predef, aPropList);

  *aEl = apply;
  NS_ADDREF(*aEl);

  return NS_OK;
}

static nsresult
textrep_MakeOtherwiseElement
(
 mathml_dom_IMathMLElement* aExpression,
 nsCStringArray& aPropList, 
 dom_IDocument* aDocument,
 mathml_dom_IMathMLContentElement** aEl
)
{
  MAKE_MATHML_OBJECT(otherwise, MathMLContentElement, "otherwise");
  nsCOMPtr<dom_INode> tmp;
  nsresult ret = 
    otherwise->AppendChild(aExpression, getter_AddRefs(tmp));
  NS_ENSURE_SUCCESS(ret, ret);

  // Put attributes onto the apply...
  textrep_ApplyPropertyList(otherwise, patt_contentelement, aPropList);

  *aEl = otherwise;

  NS_ADDREF(*aEl);
  return NS_OK;
}

static nsresult
textrep_MakePredefinedWrapElement
(
 const nsString& aType,
 mathml_dom_IMathMLElement* aExpression,
 nsCStringArray& aPropList, 
 dom_IDocument* aDocument,
 mathml_dom_IMathMLContentElement** aEl
)
{
  nsCOMPtr<dom_IElement> wrapDOMEl;
  MAKE_MATHML_OBJECT(apply, MathMLApplyElement, "apply");
  nsresult ret = aDocument->CreateElementNS(MATHML_NS, aType.get(),
                                            getter_AddRefs(wrapDOMEl));
  NS_ENSURE_SUCCESS(ret, ret);
  nsCOMPtr<mathml_dom_IMathMLPredefinedSymbol> wrapEl
    (do_QueryInterface(wrapDOMEl, &ret));
  NS_ENSURE_SUCCESS(ret, ret);

  nsCOMPtr<dom_INode> tmp;
  ret = apply->AppendChild(wrapDOMEl, getter_AddRefs(tmp));
  ret = apply->AppendChild(aExpression, getter_AddRefs(tmp));
  NS_ENSURE_SUCCESS(ret, ret);

  // Put attributes onto the apply...
  textrep_ApplyPropertyList(wrapEl, patt_predef, aPropList);

  *aEl = apply;

  NS_ADDREF(*aEl);
  return NS_OK;
}

static nsresult
textrep_DoEquality
(
 mathml_dom_IMathMLContentElement* aLHS,
 mathml_dom_IMathMLContentElement* aRHS,
 nsCStringArray& aPropList,
 dom_IDocument* aDocument,
 mathml_dom_IMathMLContentElement** aEl
)
{
  MAKE_MATHML_OBJECT(apply, MathMLApplyElement, "apply");
  MAKE_MATHML_OBJECT(eq, MathMLPredefinedSymbol, "eq");
  nsCOMPtr<dom_INode> tmp;
  nsresult ret = apply->AppendChild(eq, getter_AddRefs(tmp));
  NS_ENSURE_SUCCESS(ret, ret);
  
  textrep_ApplyPropertyList(apply, patt_contentelement, aPropList);

  nsCOMPtr<dom_INode> n;
  apply->AppendChild(aLHS, getter_AddRefs(n));
  apply->AppendChild(aRHS, getter_AddRefs(n));

  *aEl = apply;
  NS_ADDREF(*aEl);

  return NS_OK;
}

static nsresult
textrep_MakePiecewiseElement
(
 nsCOMArray<mathml_dom_IMathMLContentElement>& aCases,
 mathml_dom_IMathMLContentElement* aOtherwise,
 nsCStringArray& aPropList,
 dom_IDocument* aDocument,
 mathml_dom_IMathMLContentElement** aEl
)
{
  MAKE_MATHML_OBJECT(piecewise, MathMLPiecewiseElement, "piecewise");
  PRInt32 i;
  nsCOMPtr<dom_INode> caseEl;
  for (i = 0; i < aCases.Count(); i++)
  {
    piecewise->AppendChild(aCases[i], getter_AddRefs(caseEl));
  }
  if (aOtherwise != nsnull)
    piecewise->AppendChild(aOtherwise, getter_AddRefs(caseEl));
  *aEl = piecewise;
  NS_ADDREF(*aEl);
  return NS_OK;
}

#define CHILD_BVAR      0x0001
#define CHILD_LOWLIMIT  0x0002
#define CHILD_UPLIMIT   0x0004
#define CHILD_DEGREE    0x0008
#define CHILD_LOGBASE   0x0010

static struct FuncInfo {
  const char* match;
  enum
  {
    TOPLEVEL,
    APPLY
  } InterpretAs;
  int SpecialQualifiers;
  const char* el;
} FunctionInformation[] =
{
  {"bvar", FuncInfo::TOPLEVEL, 0, "bvar"},
  {"conditon", FuncInfo::TOPLEVEL, 0, "condition"},
  {"definite_integral", FuncInfo::APPLY,
   CHILD_BVAR|CHILD_LOWLIMIT|CHILD_UPLIMIT, "int"},
  {"degree", FuncInfo::TOPLEVEL, 0, "degree"},
  {"domainofapplication", FuncInfo::TOPLEVEL, 0, "domainofapplication"},
  {"fn", FuncInfo::TOPLEVEL, 0, "fn"},
  {"interval", FuncInfo::TOPLEVEL, 0, "interval"},
  {"lambda", FuncInfo::TOPLEVEL, 0, "lambda"},
  {"list", FuncInfo::TOPLEVEL, 0, "list"},
  {"log", FuncInfo::APPLY, CHILD_LOGBASE, "log"},
  {"logbase", FuncInfo::TOPLEVEL, 0, "logbase"},
  {"lowlimit", FuncInfo::TOPLEVEL, 0, "lowlimit"},
  {"matrix", FuncInfo::TOPLEVEL, 0, "matrix"},
  {"momentabout", FuncInfo::TOPLEVEL, 0, "momentabout"},
  {"row", FuncInfo::TOPLEVEL, 0, "matrixrow"},
  {"set", FuncInfo::TOPLEVEL, 0, "set"},
  {"uplimit", FuncInfo::TOPLEVEL, 0, "uplimit"},
  {"vector", FuncInfo::TOPLEVEL, 0, "vector"}
};

#define FUNCTION_TABLE_SIZE (sizeof(FunctionInformation)/sizeof(FunctionInformation[0]))

static FuncInfo*
GetFunctionInformation(const nsACString& name)
{
  const char* sn = name.BeginReading();
  PRUint32 upper = FUNCTION_TABLE_SIZE - 1, lower = 0;
  FuncInfo* ret = NULL;
  do
  {
    PRUint32 mid = (upper + lower) / 2;
    int cmp = strcmp(FunctionInformation[mid].match, sn);
    if (cmp == 0)
    {
      ret = FunctionInformation + mid;
      break;
    }
    else if (upper == lower)
      break;
    else if (cmp < 0)
      lower = mid + 1;
    else
    {
      if (mid == lower)
        break;
      upper = mid - 1;
    }
  }
  while (true);
  return ret;
}

struct Qualifier {
  int flag;
  const char* localname;
} Qualifiers[] =
{
  {CHILD_BVAR, "bvar"},
  {CHILD_LOWLIMIT, "lowlimit"},
  {CHILD_UPLIMIT, "uplimit"},
  {CHILD_DEGREE, "degree"},
  {CHILD_LOGBASE, "logbase"}
};
#define QUALIFIER_TABLE_SIZE (sizeof(Qualifiers)/sizeof(Qualifiers[0]))

static nsresult
textrep_DoFunction
(
 nsCString& aName,
 nsCOMArray<mathml_dom_IMathMLContentElement>& aMathList,
 dom_IDocument* aDocument,
 nsCStringArray& aPropList,
 mathml_dom_IMathMLContentElement** aEl
)
{
  FuncInfo* fi = GetFunctionInformation(aName);
  FuncInfo deffi;
  nsString el;
  if (fi == nsnull)
  {
    fi = &deffi;
    deffi.InterpretAs = FuncInfo::APPLY;
    deffi.SpecialQualifiers = 0;
    NS_ConvertUTF8toUTF16 elc(aName);
    el = elc;
  }
  else
  {
    NS_ConvertUTF8toUTF16 elc(fi->el);
    el = elc;
  }

  nsCOMPtr<dom_IElement> namedElN;
  nsresult ret =
    aDocument->CreateElementNS(MATHML_NS, el.get(),
                               getter_AddRefs(namedElN));
  NS_ENSURE_SUCCESS(ret, ret);
  nsCOMPtr<mathml_dom_IMathMLContentElement> namedEl
    (do_QueryInterface(namedElN, &ret));
  NS_ENSURE_SUCCESS(ret, ret);

  nsCOMPtr<mathml_dom_IMathMLContentElement> math;
  if (fi->InterpretAs == FuncInfo::TOPLEVEL)
  {
    math = namedEl;
  }
  else
  {
    MAKE_MATHML_OBJECT(apply, MathMLContentElement, "apply");
    nsCOMPtr<dom_INode> tmp;
    apply->AppendChild(namedEl, getter_AddRefs(tmp));
    math = apply;
  }

  nsCOMArray<mathml_dom_IMathMLContentElement> endQualifiers;

  int sq = fi->SpecialQualifiers;
  PRInt32 i;
  for (i = 0; i < (PRInt32)QUALIFIER_TABLE_SIZE; i++)
  {
    PRInt32 j, l = aMathList.Count();
    NS_ConvertUTF8toUTF16 ln16(Qualifiers[i].localname);
    for (j = 0; j < l; j++)
    {
      if ((sq & Qualifiers[i].flag) == 0)
        continue;
      nsString localname;
      aMathList[j]->GetLocalName(getter_Copies(localname));
      if (!localname.Equals(ln16))
        continue;
      
      sq &= ~Qualifiers[i].flag;
      endQualifiers.AppendObject(aMathList[j]);
      aMathList.RemoveObjectAt(j);
      j--;
      l--;
    }
  }

  // Any remaining qualifiers get pushed onto the end of the list.
  for (i = QUALIFIER_TABLE_SIZE - 1; i >= 0; i--)
  {
    if ((sq & Qualifiers[i].flag) == 0)
      continue;
    /* We never allow all the arguments to be used up by qualifiers,
     * because we automatically hold one back to use as the argument, so
     * e.g. log(5) is <apply><log/><cn>5</cn></apply> not
     * <apply><log/><logbase><cn>5</cn></logbase></apply>
     */
    if (aMathList.Count() <= 2)
      break;

    NS_ConvertUTF8toUTF16 ln16(Qualifiers[i].localname);

    nsresult ret =
      aDocument->CreateElementNS(MATHML_NS, ln16.get(),
                                 getter_AddRefs(namedElN));
    NS_ENSURE_SUCCESS(ret, ret);
    namedEl = do_QueryInterface(namedEl, &ret);
    NS_ENSURE_SUCCESS(ret, ret);

    nsCOMPtr<dom_INode> ignore;
    namedEl->AppendChild(aMathList[aMathList.Count() - 1],
                         getter_AddRefs(ignore));
    aMathList.RemoveObjectAt(aMathList.Count() - 1);

    endQualifiers.AppendObject(namedEl);
  }

  nsCOMPtr<dom_INode> ignore;
  for (i = 0; i < endQualifiers.Count(); i++)
    math->AppendChild(endQualifiers[i], getter_AddRefs(ignore));
  for (i = 0; i < aMathList.Count(); i++)
    math->AppendChild(aMathList[i], getter_AddRefs(ignore));
  
  *aEl = math;
  NS_ADDREF(*aEl);

  return NS_OK;
}

static nsresult
textrep_MakeConstant
(
 double aValue,
 dom_IDocument* aDocument,
 nsCStringArray& aPropList,
 mathml_dom_IMathMLContentElement** aEl
)
{
  MAKE_MATHML_OBJECT(cn, MathMLCnElement, "cn");

  // Put attributes onto the apply...
  textrep_ApplyPropertyList(cn, patt_cn, aPropList);

  char buf[20];
  sprintf(buf, "%g", aValue);
  nsCOMPtr<dom_IText> tEl;
  NS_ConvertUTF8toUTF16 name(buf);
  nsresult ret =
    aDocument->CreateTextNode(name.get(), getter_AddRefs(tEl));
  NS_ENSURE_SUCCESS(ret, ret);
  nsCOMPtr<dom_INode> discard;
  ret = cn->InsertArgument(tEl, 1, getter_AddRefs(discard));
  NS_ENSURE_SUCCESS(ret, ret);

  *aEl = cn;
  NS_ADDREF(*aEl);

  return NS_OK;
}

static nsresult
textrep_DoDerivative
(
 const nsString& aDifftype,
 mathml_dom_IMathMLContentElement* aExpr,
 mathml_dom_IMathMLContentElement* aBvar,
 nsCStringArray& aPropList,
 dom_IDocument* aDocument,
 mathml_dom_IMathMLContentElement** aEl
)
{
  nsCOMPtr<dom_IElement> opN;
  nsresult ret =
    aDocument->CreateElementNS(MATHML_NS, aDifftype.get(),
                               getter_AddRefs(opN));
  NS_ENSURE_SUCCESS(ret, ret);
  nsCOMPtr<mathml_dom_IMathMLContentElement> op(do_QueryInterface(opN, &ret));
  NS_ENSURE_SUCCESS(ret, ret);

  MAKE_MATHML_OBJECT(apply, MathMLApplyElement, "apply");
  nsCOMPtr<dom_INode> tmp;
  ret = apply->AppendChild(op, getter_AddRefs(tmp));
  NS_ENSURE_SUCCESS(ret, ret);

  nsCOMPtr<mathml_dom_IMathMLBvarElement> bvEl(do_QueryInterface(aBvar));
  if (bvEl == nsnull)
  {
    MAKE_MATHML_OBJECT(bvar, MathMLContainer, "bvar");
    bvEl = do_QueryInterface(bvar);
    nsCOMPtr<mathml_dom_IMathMLElement> ignore;
    bvar->InsertArgument(aBvar, 1, getter_AddRefs(ignore));
  }
  
  apply->AppendChild(bvEl, getter_AddRefs(tmp));
  apply->AppendChild(aExpr, getter_AddRefs(tmp));

  *aEl = apply;
  NS_ADDREF(*aEl);

  return NS_OK;
}

#include "TextRepLexer.h"

void
textrep_error(dom_IDocument* aDoc,
              TextRepFlexLexer* aParser,
              mathml_dom_IMathMLContentElement** aOutput,
              const char* aMessage)
{
  NS_STATIC_CAST(TextRepLexer*, aParser)->LexerError(aMessage);
}

%}
%parse-param {dom_IDocument* aDocument}
%parse-param {TextRepFlexLexer* aParser}
%parse-param {mathml_dom_IMathMLContentElement** aOutput}
%lex-param   {TextRepFlexLexer* aParser}

%token IDENTIFIER
%token QUOTED NUMBER
%token PIECEWISE CASE THEN ELSE
%left '|'
%left '&'
%left T_EQEQ T_NOTEQ T_GREATEREQUAL T_LESSEQUAL '<' '>'
%left '+' '-'
%left '*' '/'
%left '!'
%left T_DIFF T_PARTIALDIFF
%%

valid: textrep {
  *aOutput = $$.mMath;
  NS_ADDREF(*aOutput);
}

textrep: expr '=' attrs expr {
  textrep_DoEquality($1.mMath, $4.mMath, $3.mPropertyList, aDocument,
                     getter_AddRefs($$.mMath));
} | textrep '=' attrs expr {
  nsCOMPtr<dom_INode> rn;
  $1.mMath->AppendChild($1.mMath, getter_AddRefs(rn));
  $$.mMath = $1.mMath;
};
attrs: '{' attrlist '}' {
  $$.mPropertyList = $2.mPropertyList;
} | /* empty */ {
  $$.mPropertyList.Clear();
};
attrlist: attrlist attr {
  $$.mPropertyList = $1.mPropertyList;
  $$.mPropertyList.AppendCString(*$2.mPropertyList[0]);
  $$.mPropertyList.AppendCString(*$2.mPropertyList[1]);
} | /* empty */ {
  $$.mPropertyList.Clear()
};
attr: IDENTIFIER '=' QUOTED {
  $$.mPropertyList.AppendCString($1.mStr);
  $$.mPropertyList.AppendCString($3.mStr);
};
arglist: arglist ',' expr {
  PRInt32 i;
  for (i = 0; i < $1.mMathList.Count(); i++)
    $$.mMathList.AppendObject($1.mMathList[i]);
  $$.mMathList.AppendObject($3.mMath);
  $$.mPropertyList.Clear();
} | expr {
  $$.mMathList.Clear();
  $$.mMathList.AppendObject($1.mMath);
  $$.mPropertyList.Clear();
};
idform: '(' arglist ')' {
  PRInt32 i;
  for (i = 0; i < $2.mMathList.Count(); i++)
    $$.mMathList.AppendObject($2.mMathList[i]);
  $$.mOverrideBuiltin = PR_TRUE;
} | /* empty */ {
  $$.mOverrideBuiltin = PR_FALSE;
};

expr: IDENTIFIER attrs idform {
    if ($3.mOverrideBuiltin)
    {
      nsresult ret = textrep_DoFunction($1.mStr, $3.mMathList, aDocument,
                                        $2.mPropertyList,
                                        getter_AddRefs($$.mMath));
      NS_ENSURE_SUCCESS(ret,1);
    }
    else
    {
      nsresult ret =
        textrep_MakeCIOrConstant($1.mStr, $1.mOverrideBuiltin, $2.mPropertyList,
                                 aDocument, getter_AddRefs($$.mMath));
      NS_ENSURE_SUCCESS(ret,1);
    }
    $$.mPropertyList.Clear();
  } | '(' expr ')' {
    $$.mMath = $2.mMath;
    $$.mPropertyList.Clear();
  } | expr '*' attrs expr {
    nsresult ret = textrep_DoInorderExpression(NS_LITERAL_STRING("times"), $1,
                                               $4, aDocument,
                                               $3.mPropertyList,
                                               getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret,1);
    $$.mPropertyList.Clear();
  } | expr '+' attrs expr {
    nsresult ret = textrep_DoInorderExpression(NS_LITERAL_STRING("plus"), $1,
                                               $4, aDocument,
                                               $3.mPropertyList,
                                               getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret,1);
    $$.mPropertyList.Clear();
  } | expr '-' attrs expr {
    nsresult ret = textrep_DoInorderExpression(NS_LITERAL_STRING("minus"), $1,
                                               $4, aDocument,
                                               $3.mPropertyList,
                                               getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret,1);
    $$.mPropertyList.Clear();
  } | expr T_EQEQ attrs expr {
    nsresult ret = textrep_DoInorderExpression(NS_LITERAL_STRING("eq"), $1,
                                               $4, aDocument,
                                               $3.mPropertyList,
                                               getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret, 1);
    $$.mPropertyList.Clear();
  } | expr T_NOTEQ attrs expr {
    nsresult ret = textrep_DoInorderExpression(NS_LITERAL_STRING("neq"), $1,
                                               $4, aDocument,
                                               $3.mPropertyList,
                                               getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret,1);
    $$.mPropertyList.Clear();
  } | expr '<' attrs expr {
    nsresult ret = textrep_DoInorderExpression(NS_LITERAL_STRING("lt"), $1,
                                               $4, aDocument,
                                               $3.mPropertyList,
                                               getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret,1);
    $$.mPropertyList.Clear();
  } | expr '>' attrs expr {
    nsresult ret = textrep_DoInorderExpression(NS_LITERAL_STRING("gt"), $1,
                                               $4, aDocument,
                                               $3.mPropertyList,
                                               getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret,1);
    $$.mPropertyList.Clear();
  } | expr T_GREATEREQUAL attrs expr {
    nsresult ret = textrep_DoInorderExpression(NS_LITERAL_STRING("geq"), $1,
                                               $4, aDocument,
                                               $3.mPropertyList,
                                               getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret,1);
    $$.mPropertyList.Clear();
  }| expr T_LESSEQUAL attrs expr {
    nsresult ret = textrep_DoInorderExpression(NS_LITERAL_STRING("leq"), $1,
                                               $4, aDocument,
                                               $3.mPropertyList,
                                               getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret,1);
    $$.mPropertyList.Clear();
  } | expr '/' attrs expr {
    nsresult ret = textrep_DoInorderExpression(NS_LITERAL_STRING("divide"), $1,
                                               $4, aDocument,
                                               $3.mPropertyList,
                                               getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret,1);
    $$.mPropertyList.Clear();
  } | expr '|' attrs expr {
    nsresult ret = textrep_DoInorderExpression(NS_LITERAL_STRING("or"), $1,
                                               $4, aDocument,
                                               $3.mPropertyList,
                                               getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret,1);
    $$.mPropertyList.Clear();
  } | expr '&' attrs expr {
    nsresult ret = textrep_DoInorderExpression(NS_LITERAL_STRING("and"), $1,
                                               $4, aDocument,
                                               $3.mPropertyList,
                                               getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret,1);
    $$.mPropertyList.Clear();
  } | NUMBER attrs {
    nsresult ret =
      textrep_MakeConstant($1.mNumber, aDocument, $2.mPropertyList,
                           getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret,1);
    $$.mPropertyList.Clear();
  } | PIECEWISE attrs '(' caselist maybeelse ')' {
    nsresult ret =
      textrep_MakePiecewiseElement($4.mMathList, $5.mMath, $2.mPropertyList,
                                   aDocument, getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret, 1);
    $$.mPropertyList.Clear();
  } | '!' attrs expr {
    nsresult ret =
      textrep_MakePredefinedWrapElement(NS_LITERAL_STRING("not"), $3.mMath,
                                        $2.mPropertyList, aDocument,
                                        getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret, 1);
    $$.mPropertyList.Clear();
  } | '-' attrs expr %prec '!' {
    nsresult ret =
      textrep_MakePredefinedWrapElement(NS_LITERAL_STRING("minus"), $3.mMath,
                                        $2.mPropertyList, aDocument,
                                        getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret, 1);
    $$.mPropertyList.Clear();
  } | T_DIFF expr ')' '/' T_DIFF expr ')' attrs {
    nsresult ret =
      textrep_DoDerivative(NS_LITERAL_STRING("diff"), $2.mMath,
                           $6.mMath, $8.mPropertyList, aDocument,
                           getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret, 1);
    $$.mPropertyList.Clear();
  } | T_PARTIALDIFF expr ')' '/' T_PARTIALDIFF expr ')' attrs {
    nsresult ret =
      textrep_DoDerivative(NS_LITERAL_STRING("partialdiff"), $2.mMath,
                           $6.mMath, $8.mPropertyList, aDocument,
                           getter_AddRefs($$.mMath));
    NS_ENSURE_SUCCESS(ret, 1);
    $$.mPropertyList.Clear();
  };
caselist: caselist casepair {
    PRInt32 i;
    $$.mMathList.Clear();
    for (i = 0; i < $1.mMathList.Count(); i++)
      $$.mMathList.AppendObject($1.mMathList[i]);
    $$.mMathList.AppendObject($2.mMath);
} | /* empty */ {
  $$.mMathList.Clear();
};
casepair: CASE attrs expr THEN expr {
  nsresult ret =
    textrep_MakeCaseElement($3.mMath, $5.mMath, aDocument, $2.mPropertyList,
                            getter_AddRefs($$.mMath));
  NS_ENSURE_SUCCESS(ret,1);
};
maybeelse: ELSE attrs expr {
  nsresult ret =
    textrep_MakeOtherwiseElement($3.mMath, $2.mPropertyList, aDocument,
                                 getter_AddRefs($$.mMath));
  NS_ENSURE_SUCCESS(ret,1);
} | /* empty */ {
  $$.mMath = nsnull;
};
